from json import load
from unittest.mock import patch

from pandas import Series
from pandas.testing import assert_series_equal

# noinspection PyProtectedMember
from fipiran import FundProfile, _core, funds, search, Symbol
from fipiran._core import YK, jdatetime, DataFrame

disable_api = patch.object(_core, 'api', side_effect=RuntimeError(
    'api should not be called during tests'))


def setup_module():
    disable_api.start()


def teardown_module():
    disable_api.stop()


def patch_api(name):
    with open(f'{__file__}/../testdata/{name}.json', 'rb') as f:
        j = load(f)
    return patch.object(_core, 'api', lambda _: j)


def patch_fipiran(name):
    with open(f'{__file__}/../testdata/{name}.html', 'r', encoding='utf8') as f:
        text = f.read()
    return patch.object(_core, 'fipiran', lambda _: text.translate(YK))


fp = FundProfile(11215)


def test_repr():
    assert repr(fp) == 'FundProfile(11215)'
    assert repr(FundProfile('11215')) == "FundProfile('11215')"


@patch_api('getfundchartasset_atlas')
def test_asset_allocation():
    d = fp.asset_allocation()
    del d['fiveBest']
    assert sum(d.values()) == 100


@patch_api('getfundchart_atlas')
def test_issue_cancel_history():
    df = fp.issue_cancel_history()
    assert_series_equal(df.dtypes, Series(['float64', 'float64'], ['issueNav', 'cancelNav']))
    assert len(df) == 366
    assert df.index.dtype == '<M8[ns]'


@patch_api('getfundnetassetchart_atlas')
def test_nav_history():
    df = fp.nav_history()
    assert_series_equal(df.dtypes, Series(['int64'], ['netAsset']))
    assert len(df) == 366
    assert df.index.dtype == '<M8[ns]'


@patch_api('getfund_atlas')
def test_info():
    info = fp.info()
    assert len(info) == 54
    assert type(info) is dict


@patch_api('fundlist')
def test_funds():
    df = funds()
    assert len(df) == 272


@patch_api('autocomplete_arzesh')
def test_search():
    assert search('ارزش') == [
        {'LVal18AFC': 'وآفر', 'LSoc30': ' سرمايه گذاري ارزش آفرينان'},
        {'LVal18AFC': 'وارزش', 'LSoc30': 'ارزش آفرينان پاسارگاد'},
        {'LVal18AFC': 'وآفر', 'LSoc30': 'سرمايه گذاري ارزش آفرينان'},
        {'LVal18AFC': 'ارزش', 'LSoc30': 'صندوق س ارزش آفرين بيدار-سهام'},
        {'LVal18AFC': 'ومدير', 'LSoc30': 'گ.مديريت ارزش سرمايه ص ب كشوري'}]


@patch_fipiran('SymbolFMelli')
def test_symbol_from_name():
    s = Symbol.from_name('فملی')
    assert f'{s!r}' == "Symbol(35425587644337450, 'فملی')"


@patch_fipiran('priceDataFMelli')
def test_symbol_price_data():
    s = Symbol(35425587644337450, 'فملی')
    assert s.price_data() == {
        'PriceMin': 13140.0, 'PriceMax': 13770.0, 'PDrCotVal': 13290.0,
        'PriceFirst': 13770.0, 'PClosing': 13300.0, 'changepdr': -1.77,
        'changepc': -1.7, 'prevPrice': 13530.0, 'ZTotTran': 10103.0,
        'QTotTran5J': 71934487.0, 'QTotCap': 956433812420.0,
        'Deven': jdatetime(1400, 8, 1, 12, 30), 'tmin': 12860.0, 'tmax': 14200.0}


@patch_fipiran('BestLimitDataFMelli')
def test_symbol_best_limit_data():
    d1, d2 = Symbol(35425587644337450, 'فملی').best_limit_data()
    assert type(d1) is type(d2) is DataFrame


@patch_fipiran('RefrenceDataFmelli')
def test_symbol_refrence_data():
    assert Symbol(35425587644337450, 'فملی').refrence_data() == {
        'نام نماد': 'فملی',
        'نام شرکت': 'ملی\u200c صنایع\u200c مس\u200c ایران\u200c\u200c',
        'نام صنعت': 'فلزات اساسی', 'وضعیت': 'مجاز',
        'بازار': 'بورس-بازاراول-تابلو اصلی',
        'حجم مبنا': '8869180', 'کد معاملاتی نماد': 'IRO1MSMI0001'}


@patch_fipiran('statistic30Fmelli')
def test_symbol_refrence_data():
    assert type(Symbol(35425587644337450, 'فملی').statistic(30)) is DataFrame



from __future__ import annotations

from functools import partial
from typing import Literal

from requests import get
from pandas import DataFrame, to_datetime, read_html
from bs4 import BeautifulSoup
from jdatetime import datetime as jdatetime


API = 'https://fund.fipiran.ir/api/v1/'
FIPIRAN = 'https://www.fipiran.com/'
YK = ''.maketrans('يك', 'یک')


class FundProfile:
    __slots__ = 'reg_no'

    def __init__(self, reg_no: int | str):
        self.reg_no = reg_no

    def __repr__(self):
        return f'{type(self).__name__}({self.reg_no!r})'

    def asset_allocation(self) -> dict:
        """Return a dict where values are percentage of each kind of asset."""
        return api(f'chart/getfundchartasset?regno={self.reg_no}')

    def issue_cancel_history(self) -> DataFrame:
        j = api(f'chart/getfundchart?regno={self.reg_no}')
        df = DataFrame(j)
        df['date'] = to_datetime(df['date'])
        df.set_index('date', inplace=True)
        return df

    def nav_history(self) -> DataFrame:
        j = api(f'chart/getfundnetassetchart?regno={self.reg_no}')
        df = DataFrame(j)
        df['date'] = to_datetime(df['date'])
        df.set_index('date', inplace=True)
        return df

    def info(self):
        return api(f'fund/getfund?regno={self.reg_no}')['item']


class Symbol:
    __slots__ = 'inscode', 'name'

    def __init__(self, inscode: int | str, name: str):
        """Use `from_name` or `from_inscode` if only 1 parameter is known."""
        self.inscode = inscode
        self.name = name

    def __repr__(self):
        return f'{type(self).__name__}({self.inscode!r}, {self.name!r})'

    @staticmethod
    def from_name(symbolpara: str, /) -> Symbol:
        text = fipiran(f'Symbol?symbolpara={symbolpara}')
        start = text.find("'inscode': '") + 12
        end = text.find("'", start)
        inscode = int(text[start: end])
        start = text.find("var symbolpara = '") + 18
        end = text.find("'", start)
        name = text[start: end]
        return Symbol(inscode, name)

    def price_data(self) -> dict:
        # note: some fields like Trailing P/E and Forward P/E are *currently*
        # not computed by fipiran and are always empty.
        text = fipiran(f'Symbol/_priceData?inscode={self.inscode}')
        bs = soup(text)
        so = bs.select_one
        d = {}

        def num(s: str) -> float:
            s = s.replace(',', '').strip(' ')
            if s[0] == '(':  # negative value
                return -float(s.strip('()'))
            return float(s)

        for k in (  # numerical values
            'PriceMin', 'PriceMax', 'PDrCotVal', 'PriceFirst',
            'PClosing', 'changepdr', 'changepc', 'prevPrice', 'ZTotTran',
            'QTotTran5J', 'QTotCap'
        ):
            d[k] = num(so(f'#{k}').text)

        d['Deven'] = jdatetime.strptime(so('#Deven').text, '%Y/%m/%d-%H:%M:%S')

        tmin, tmax = bs.find(string='قیمت مجاز').next.text.split(' - ')
        d['tmin'] = num(tmin)
        d['tmax'] = num(tmax)

        return d

    def best_limit_data(self) -> list[DataFrame]:
        text = fipiran(f'Symbol/_priceData?inscode={self.inscode}')
        return read_html(text)

    def refrence_data(self) -> dict:
        text = fipiran(f'Symbol/_RefrenceData?symbolpara={self.name}')
        bs = soup(text)
        h4s = [i.text.strip(': ') for i in bs.select('h4')]
        spans = [i.text for i in bs.select('span')]
        d = dict(zip(h4s, spans))
        # e.g. 'کد معاملاتی نماد', 'IRO1MSMI0001'
        k, v = bs.select_one('span')['title'].split(' :')
        d[k] = v
        return d

    def statistic(self, days: Literal[365, 180, 90, 30, 7]) -> DataFrame:
        return read_html(
            fipiran(f'Symbol/statistic{days}?inscode={self.inscode}'))[0]


soup = partial(BeautifulSoup, features='lxml')


def fipiran(path: str) -> str:
    return get(f'{FIPIRAN}{path}').content.decode().translate(YK)


def api(path) -> dict | list:
    return get(API + path).json()


def funds() -> DataFrame:
    return DataFrame(api('fund/fundlist')['items'])


def search(term) -> list[dict]:
    return get(FIPIRAN + 'Home/AutoComplete', data=(('term', term),)).json()
